"""The SoCo-CLI API.

Provides a few simple, high-level functions that allow the features of SoCo-CLI
to be used in other programs.
"""

import logging
import sys
from io import StringIO
from signal import SIGINT, signal

from soco import SoCo

from soco_cli.action_processor import process_action
from soco_cli.speakers import Speakers
from soco_cli.utils import (
    configure_logging,
    create_speaker_cache,
    get_speaker,
    set_api,
    set_speaker_list,
    sig_handler,
    speaker_cache,
)


def run_command(speaker_name, action, *args, use_local_speaker_list=False):
    """Use SoCo-CLI to run a sonos command.

    The exit code, output string and error message string are returned as a
    three-tuple. If the exit code is non-zero, the error message will be
    populated and the output string will always be empty.

    All exceptions are caught when this function is run. Exception details
    will be returned in the error message string.

    Args:
        speaker_name (str): The name of the speaker, or its IP address
        action (str): The The name of the SoCo-CLI action to perform
        *args (list[str]): The set of arguments that accompany the action

    Returns:
        int, str, str: a three-tuple of exit_code, output_string and
        error_msg.
    """

    # Prevent errors from causing exit
    set_api()

    # Capture stdout and stderr for the duration of this command
    output = StringIO()
    sys.stdout = output
    error = StringIO()
    sys.stderr = error

    speaker = None
    exception_error = None

    # Can pass a SoCo object instead of the speaker name (not documented)
    if isinstance(speaker_name, SoCo):
        speaker = speaker_name

    elif isinstance(speaker_name, str):
        try:
            speaker = _get_soco_object(
                speaker_name, use_local_speaker_list=use_local_speaker_list
            )
        except Exception as e:
            logging.info("Exception: {}".format(e))
            exception_error = e

    return_value = False

    if speaker:
        try:
            return_value = process_action(
                speaker, action, args, use_local_speaker_list=use_local_speaker_list
            )
        except Exception as e:
            logging.info("Exception: {}".format(e))
            exception_error = e

        output_msg = output.getvalue().rstrip()
        error_out = error.getvalue().rstrip()

        if not output_msg == "":
            lines = output_msg.splitlines()
            if len(lines) > 1 and lines[0] != "":
                output_msg = "\n" + output_msg
            if len(lines) > 1 and output_msg[len(lines) - 1] != "":
                output_msg = output_msg + "\n"

        if exception_error:
            if error_out:
                error_out = error_out + "\nError: " + str(exception_error)
            else:
                error_out = "Error: " + str(exception_error)

        if not return_value:
            if error_out == "":
                hint = " ... missing spaces around ':'?" if ":" in action else ""
                error_out = "Error: Action '{}' not found{}".format(action, hint)
            return_value = (1, output_msg, error_out)
        else:
            return_value = (0, output_msg, error_out)
    else:
        return_value = (
            1,
            "",
            "Speaker '{}' not found: {}".format(speaker_name, exception_error),
        )

    # Restore stdout and stderr
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__

    logging.info("Return value: {}".format(return_value))

    return return_value


def set_log_level(log_level="None"):
    """Convenience function to set up logging.

    Args:
        log_level (str): Can be one of None, Critical, Error, Warn, Info, Debug.
    """
    configure_logging(log_level)


def handle_sigint():
    """Convenience function to set up a more graceful CTRL-C (sigint) handler."""
    signal(SIGINT, sig_handler)


def rescan_speakers():
    """Run full network scan to find speakers."""
    _check_for_speaker_cache()
    speaker_cache().scan(reset=True)


def rediscover_speakers():
    """Run normal SoCo discovery to discover speakers."""
    _check_for_speaker_cache()
    speaker_cache().discover(reset=True)


def get_all_speakers():
    """Return all SoCo instances."""
    _check_for_speaker_cache()
    return [s[0] for s in speaker_cache().get_all_speakers()]


def get_all_speaker_names():
    """Return all speaker names."""
    _check_for_speaker_cache()
    return speaker_cache().get_all_speaker_names()


def get_soco_object(speaker_name, use_local_speaker_list=False):
    """Uses the full set of soco_cli strategies to find a speaker.

    Args:
        speaker_name (str): The name of the speaker to find.
        use_local_speaker_list (bool): Whether to use the local speaker cache.

    Returns:
        (SoCo, str): Tuple of SoCo object, or None if no speaker is found,
        and an error message.
    """
    set_api()

    error = StringIO()
    sys.stderr = error

    speaker = _get_soco_object(speaker_name, use_local_speaker_list)

    sys.stderr = sys.__stderr__

    error_msg = error.getvalue().rstrip()
    if not speaker and error_msg == "":
        error_msg = "Speaker not found"

    return speaker, error_msg


def _get_soco_object(speaker_name, use_local_speaker_list=False):
    """Internal helper version that doesn't redirect stderr."""

    if use_local_speaker_list:
        _setup_local_speaker_list()

    _check_for_speaker_cache()

    return get_speaker(speaker_name, use_local_speaker_list)


def _check_for_speaker_cache():
    if not speaker_cache():
        create_speaker_cache(max_threads=256, scan_timeout=1.0, min_netmask=24)


# For local speaker list operations
speaker_list_set = False


def _setup_local_speaker_list():
    global speaker_list_set
    if not speaker_list_set:
        speaker_list = Speakers()
        if not speaker_list.load():
            logging.info("Start speaker discovery")
            speaker_list.discover()
            speaker_list.save()
        set_speaker_list(speaker_list)
    speaker_list_set = True



"""Parse M3U files."""

# Derived from https://github.com/dvndrsn/M3uParser ... thanks!
#
# more info on the M3U file format available here:
# http://n4k3d.com/the-m3u-file-format/

from soco_cli.utils import error_and_exit


class Track:
    def __init__(self, length, title, path):
        self.length = length
        self.title = title
        self.path = path


"""
    song info lines are formatted like:
    EXTINF:419,Alice In Chains - Rotten Apple
    length (seconds)
    Song title
    file name - relative or absolute path of file
"""


def parse_m3u(m3u_file):
    with open(m3u_file, "r") as infile:
        """
        Parse file contents. Files with an M3U/M3U8 extension must follow conventions.
        """

        if m3u_file.lower().endswith(".m3u") or m3u_file.lower().endswith(".m3u8"):
            line = infile.readline()
            if not line.startswith("#EXTM3U"):
                error_and_exit(
                    "File '{}' lacks '#EXTM3U' as first line".format(m3u_file)
                )
                return None

        playlist = []
        song = Track(None, None, None)
        for line in infile:
            line = line.strip()
            if line.startswith("#EXTINF:"):
                # pull length and title from #EXTINF line
                length, title = line.split("#EXTINF:")[1].split(",", 1)
                song = Track(length, title, None)
            elif line.startswith("#"):
                # Comment line
                pass
            elif len(line) != 0:
                # pull song path from all other, non-blank lines
                song.path = line
                playlist.append(song)
                # reset the song variable so it doesn't use the same EXTINF more than once
                song = Track(None, None, None)

        return playlist



"""Manages speaker information for Cached Discovery mode."""

import ipaddress
import logging
import os
import pickle
from collections import namedtuple

import soco
import tabulate

from soco_cli.match_speaker_names import speaker_name_matches

# Type for holding speaker details
SonosDevice = namedtuple(
    "SonosDevice",
    [
        "household_id",
        "ip_address",
        "speaker_name",
        "is_visible",
        "model_name",
        "display_version",
    ],
    rename=False,
)


class Speakers:
    """A class for discovering Sonos speakers, saving and loading speaker data,
    and finding speakers by name. An alternative to using SoCo discovery.
    """

    def __init__(
        self,
        save_directory=None,
        save_file=None,
        network_threads=256,
        network_timeout=0.1,
        min_netmask=24,
    ):
        self._save_directory = (
            save_directory
            if save_directory
            else os.path.expanduser("~") + "/.soco-cli/"
        )
        self._save_file = save_file if save_file else "speakers_v2.pickle"
        self.remove_deprecated_pickle_files()
        self._network_threads = network_threads
        self._network_timeout = network_timeout
        self._min_netmask = min_netmask
        self._speakers = []

    def remove_deprecated_pickle_files(self):
        """Remove any older, incompatible versions of the pickle file"""
        for old_file in ["speakers.pickle"]:
            pathname = self._save_directory + old_file
            if os.path.exists(pathname):
                logging.info("Removing old local speaker cache {}".format(pathname))
                # print("Removing deprecated local speaker file:", pathname)
                os.remove(pathname)

    @property
    def speaker_cache_loaded(self):
        if self._speakers:
            return True
        else:
            return False

    @property
    def speaker_cache_file_exists(self):
        if os.path.exists(self.save_pathname):
            return True
        else:
            return False

    @property
    def speakers(self):
        return self._speakers

    @property
    def save_directory(self):
        return self._save_directory

    @save_directory.setter
    def save_directory(self, directory):
        self._save_directory = directory

    @property
    def save_file(self):
        return self._save_file

    @save_file.setter
    def save_file(self, file):
        self._save_file = file

    @property
    def save_pathname(self):
        return self._save_directory + self._save_file

    @property
    def network_threads(self):
        return self._network_threads

    @network_threads.setter
    def network_threads(self, threads):
        self._network_threads = threads

    @property
    def network_timeout(self):
        return self._network_timeout

    @network_timeout.setter
    def network_timeout(self, timeout):
        self._network_timeout = timeout

    @property
    def min_netmask(self):
        return self._min_netmask

    @min_netmask.setter
    def min_netmask(self, min_netmask):
        self._min_netmask = min_netmask

    def save(self):
        """Saves the speaker list as a pickle file."""
        if self._speakers:
            if not os.path.exists(self._save_directory):
                os.mkdir(self._save_directory)
            with open(self.save_pathname, "wb") as f:
                pickle.dump(self._speakers, f)
            return True
        else:
            return False

    def load(self):
        """Loads a saved speaker list"""
        if os.path.exists(self.save_pathname):
            try:
                with open(self.save_pathname, "rb") as f:
                    self._speakers = pickle.load(f)
            except:
                return False
            return True
        else:
            return False

    def clear(self):
        """Clears the in-memory speaker list"""
        self._speakers = []

    def remove_save_file(self):
        """Removes the saved speaker list file"""
        os.remove(self.save_pathname)
        return self.save_pathname

    def rename(self, old_name, new_name):
        for index, speaker in enumerate(self._speakers):
            if old_name.replace("’", "'") == speaker.speaker_name.replace("’", "'"):
                # Update old record, delete, replace with new
                new_speaker = speaker._replace(speaker_name=new_name)
                del self._speakers[index]
                self._speakers.append(new_speaker)
                logging.info(
                    "Renamed speaker in cache: '{}' to '{}'".format(old_name, new_name)
                )
                self.save()
                logging.info("Saved updated cache file")
                return True
        logging.info("Failed to find speaker '{}' for rename".format(old_name))
        return False

    @staticmethod
    def is_ipv4_address(ip_address):
        """Tests for an IPv4 address"""
        try:
            ipaddress.IPv4Network(ip_address)
            return True
        except ValueError:
            return False

    @staticmethod
    def get_sonos_device_data(ip_addr):
        """Get information from a Sonos device"""
        try:
            speaker = soco.SoCo(str(ip_addr))
            info = speaker.get_speaker_info(refresh=True)
            return SonosDevice(
                speaker.household_id,
                str(ip_addr),
                info["zone_name"],
                speaker.is_visible,
                info["model_name"],
                info["display_version"],
            )
        except:
            logging.info("Not a Sonos device: '{}'".format(ip_addr))
            return None

    def discover(self):
        """Discover the Sonos speakers on the network(s) to which
        this host is attached."""

        devices = soco.discovery.scan_network(
            include_invisible=True,
            multi_household=True,
            scan_timeout=self._network_timeout,
            max_threads=self._network_threads,
            min_netmask=self._min_netmask,
        )

        # Populate the device information for each speaker
        for device in devices:
            self._speakers.append(self.get_sonos_device_data(device.ip_address))

    def find(self, speaker_name, require_visible=True):
        """Find a speaker by name and return its SoCo object."""

        speaker_names = set()
        return_speaker = None

        for speaker in self._speakers:
            if require_visible and not speaker.is_visible:
                continue

            match, exact = speaker_name_matches(speaker_name, speaker.speaker_name)

            if match and exact:
                speaker_names.add(speaker.speaker_name)
                return soco.SoCo(speaker.ip_address)

            if match and not exact:
                speaker_names.add(speaker.speaker_name)
                if not return_speaker:
                    return_speaker = soco.SoCo(speaker.ip_address)

        if len(speaker_names) > 1:
            print(
                "Speaker name '{}' is ambiguous within {}".format(
                    speaker_name, speaker_names
                )
            )
            return None

        return return_speaker

    def get_all_speakers(self):
        soco_speakers = []
        for speaker in self._speakers:
            soco_speakers.append(soco.SoCo(speaker.ip_address))
        if soco_speakers:
            return soco_speakers
        else:
            return None

    def get_all_speaker_names(self, include_invisible=False):
        soco_speaker_names = []
        for speaker in self._speakers:
            if speaker.is_visible:
                soco_speaker_names.append(speaker.speaker_name)
        soco_speaker_names.sort()
        return soco_speaker_names

    def print(self):
        if not self._speakers:
            return
        households = {}
        num_devices = 0
        for device in self._speakers:
            if device.household_id not in households:
                households[device.household_id] = []
            if device.is_visible:
                visible = "Visible"
            else:
                visible = "Hidden"
            households[device.household_id].append(
                (
                    device.speaker_name,
                    device.ip_address,
                    device.model_name.replace("Sonos ", ""),
                    visible,
                    device.display_version,
                )
            )
            num_devices += 1

        headers = [
            "Room/Zone Name",
            "IP Address",
            "Device Model",
            "Visibility",
            "SW Version",
        ]
        for household in households:
            print()
            print("Sonos Household: {}\n".format(household))
            print(
                tabulate.tabulate(
                    sorted(households[household]), headers, numalign="left"
                )
            )
            print()

        print("{} Sonos Household(s) found".format(len(households)))
        print("{} Sonos device(s) found".format(num_devices))
        print()
